import inspect
import os
import pkgutil
import sys

from kck.lib.exceptions import PrimerComputerReturnedNoResults
from .config import kck_config_lookup
from .kck_database import KCKDatabase
from .framework_actor import FrameworkActor

SCRIPT_DIRPATH = os.path.dirname(os.path.realpath(__file__))
KCK_SOURCES_DIRPATH = os.path.join(SCRIPT_DIRPATH, "..")

QUERY_TYPE_ROWS = 1
QUERY_TYPE_DICTS = 2

def refresh_for_param_dict_list(primer_obj, param_dict_list):
    cache_obj = primer_obj.get_cache()
    for param_dict in param_dict_list:
        cache_obj.refresh(
            key="{}{}{}".format(primer_obj.name, primer_obj.keysep, primer_obj.param_dict_to_key(param_dict)),
            hints=None,
            queued=True
        )


class KCKPrimer(FrameworkActor):
    keys = None
    database = None
    expire = None
    query_template = None
    query_type = QUERY_TYPE_ROWS
    result_filter = None
    domain_key = None
    processes = {
        "refresh__over_domain": {
            "type": "key",
            "key_attribute": "domain_key",
            "hooks": {
                "pre": None,
                "value": [refresh_for_param_dict_list],
                "post": None,
            }
        }
    }

    def __str__(self):
        return "{}: [keys: {}, database: {}, parameters: {}, expire: {}, qrytmpl: {}]".format(
            self.__class__,
            self.keys,
            self.database,
            self.parameters,
            self.expire,
            self.query_template
        )

    def refresh(self, keystr):
        pass

    def compute(self, keystr):

        ret = []

        if self.query_type == QUERY_TYPE_ROWS:
            ret = [ list(x) for x in KCKDatabase.query(
                self.database,
                self.query_template,
                self.key_to_param_dict(keystr)
            ).fetchall() ]

        if self.query_type == QUERY_TYPE_DICTS:
            res = KCKDatabase.query(
                self.database,
                self.query_template,
                self.key_to_param_dict(keystr)
            )

            while True:

                row = res.fetchone()
                if row is None:
                    break
                rowdict = dict(zip(row.keys(), row.values()))


                ret.append(rowdict)

            res.close()

        if self.result_filter is not None:
            ret = self.result_filter(ret)

        if ret is None or not bool(ret):
            raise PrimerComputerReturnedNoResults
        return ret

    @classmethod
    def register_primers(cls, cache_obj):

        def proc_primer_instance(primer_name, primer_obj):
            cache_obj.register_primer(primer_name, primer_obj)

        # register the path one level up from the primers dirpath as a python library path
        above_dirpath = cls.make_relpath_absolute(
            os.path.join(
                cls.make_relpath_absolute(kck_config_lookup("kck", "primers_dirpath")),
                '..'
            )
        )
        if above_dirpath not in sys.path:
            sys.path.append(above_dirpath)

        # import the kprimers module, instantiate each primer class found and register it with the cache instance
        import kprimers
        for loader, name, ispkg in pkgutil.iter_modules(kprimers.__path__):
            m = loader.find_module(name).load_module(name)
            for n, c in inspect.getmembers(m, inspect.isclass):
                if not issubclass(c, KCKPrimer) or c.__name__ == "KCKPrimer":
                    continue
                proc_primer_instance(name, c())




import os
import datetime

from jsonpath_ng import jsonpath, parse

from .config import kck_config_lookup
from .kck_primer import KCKPrimer, QUERY_TYPE_DICTS, QUERY_TYPE_ROWS
from .yaml_reader import YAMLReader

class KCKYamlPrimer(KCKPrimer, YAMLReader):

    refresh_after_updates = []
    refresh_after_primes = []


    def __init__(self, yaml_filepath):

        self.filepath = yaml_filepath

        self._proc_primer_yaml()


    @classmethod
    def register_primers(cls, cache_obj):
        """register all yaml-based primers in the primers dir"""

        def fnfilter(filepath):
            # print("fp: {}, fpl: {}".format(filepath, filepath[-3:]))
            if filepath[-4:] == ".yml":
                return True
            return False

        def proc_primer_yaml(filepath):
            cache_obj.register_primer(
                os.path.basename(filepath)[:-4],
                KCKYamlPrimer(filepath)
            )

        cls.proc_files_in_dirpath(
            cls.make_relpath_absolute(kck_config_lookup("kck", "primers_dirpath")),
            fnfilter,
            proc_primer_yaml
        )

    def register_hooks(self):

        # register after_primes refreshes



        super(KCKYamlPrimer, self).register_hooks()

        cache_obj = self.get_cache()

        # add hooks for refresh-after-primes

        refresh_str_format = "{}{}".format(self.name, self.keysep)+"{}"

        def refresh_primer_val(key=None, hints=None, param_key=None):
            print("refresh_primer_val(key={}, hints={}, param_key={})".format(key, hints, param_key))
            if bool(hints["success"]):
                #
                # - key is the key that just got refreshed, triggering the hook
                # - param_key is the key that was listed in refresh/after_primes
                #

                # --- build the keystr to send to the cache_obj refresh() method

                # get the param dict from the refresh_str_format
                primer_obj = cache_obj.primer_obj(refresh_str_format.format(""))

                # get the param dict that describes dependent key in terms of current key
                param_dict = self.dependency_key_to_param_dict(param_key, key)

                refresh_key = refresh_str_format.format(self.param_dict_to_key(param_dict))

                self.get_cache().refresh(
                    key=refresh_key,
                    hints=hints,
                    queued=True
                )

        for key_tmpl in self.refresh_after_primes:

            # param_dict = self.dependency_key_to_param_dict(key_tmpl, key)

            # get the primer obj for the key
            target_primer_obj = cache_obj.primer_obj(key_tmpl)

            # add refresh_primer_val() as a post-prime hook
            depkey = self.dependency_key()
            target_primer_obj.add_hook("post_prime", key_tmpl, refresh_primer_val)

        for key_tmpl in self.refresh_after_updates:
            u = cache_obj.updater_obj(key_tmpl)
            u.add_hook("post_update", key_tmpl, refresh_primer_val)

    def _proc_primer_yaml(self):

        p = self.read_yaml(self.filepath)

        # the key is the same as the yaml file (minus the .yml)
        self.keys = [os.path.basename(self.filepath)[:-4]]

        self.database = p["database"]

        self.parameters = p["parameters"] if "parameters" in p else []

        self.expire = datetime.timedelta(seconds=p["expire_seconds"]) if "expire_seconds" in p else None

        self.query_template = p["query"]["template"]

        self.query_type = QUERY_TYPE_DICTS if "type" in p["query"] and p["query"]["type"] == "dicts" else QUERY_TYPE_ROWS

        if "process_result" in p["query"]:
            if "jsonpath_first" in p["query"]["process_result"]:
                jpparser = parse(p["query"]["process_result"]["jsonpath_first"])
                self.result_filter = lambda x: jpparser.find(x)[0].value if x is not None and bool(x) else None

        if "refresh" in p:
            try:
                self.refresh_after_primes = p["refresh"]["after_primes"]
            except KeyError:
                pass

            try:
                self.refresh_after_updates = p["refresh"]["after_updates"]
            except KeyError:
                pass

            try:
                self.domain_key = p["refresh"]["domain_key"]
            except KeyError:
                pass




import traceback


class BaseKCKException(Exception):

    msg = None

    def failure_info_dict(self):
        ret = dict(
            error=dict(
                name=self.__class__.__name__,
                args=self.args,
                traceback=traceback.format(self.__traceback__)
            )
        )

        if self.msg is not None:
            ret["error"]["message"] = self.msg

        return ret


class KCKKeyNotSetException(BaseKCKException):
    key = None
    tbl = None

    def __init__(self, key, tbl, msg="primer found, but compute() returned no results"):
        self.key = key
        self.tbl = tbl
        self.msg = msg


class KCKKeyNotRegistered(BaseKCKException):
    key = None

    def __init__(self, key):
        self.key = key


class UnexpectedPrimerType(BaseKCKException):
    pass


class KCKUnknownKey(BaseKCKException):
    key = None

    def __init__(self, key):
        self.key = key


class PrimerComputerReturnedNoResults(BaseKCKException):
    pass


class AuthTokenInvalid(BaseKCKException):
    msg = "Invalid authorization token"
    token = None

    def __init__(self, token):
        self.token = token


class AuthTokenExpired(BaseKCKException):
    msg = "Expire authorization token"
    token = None

    def __init__(self, token):
        self.token = token


class AuthHeaderNotPresent(BaseKCKException):
    msg = "Authorization header not present"


class AuthLoginUnknownUser(BaseKCKException):
    msg = "Unknown user or bad password"
    username = None

    def __init__(self, username):
        self.username = username


import json
import datetime
import traceback

import tornado.web

from urllib.parse import urlparse, parse_qs
from tornado.options import define, options, parse_command_line
from kck.lib.exceptions import KCKKeyNotSetException, AuthHeaderNotPresent, AuthLoginUnknownUser
from kck.lib.kck_auth import KCKAuth
from .config import kck_config_lookup

define("port", default=8001, help="run on the given port", type=int)
define("debug", default=False, help="run in debug mode")

try:
    use_auth = kck_config_lookup("kck", "enable_auth")
except KeyError:
    use_auth = False

auth_obj = KCKAuth(auth_dict=kck_config_lookup("auth")) if use_auth else KCKAuth(auth_dict=None)


def authed_handler_method(f):
    def wrapper(*args):
        self = args[0]

        if use_auth:
            auth_token = self.request.headers.get('Authorization')

            if not auth_token:
                raise AuthHeaderNotPresent

            auth_obj.authed_userid(auth_obj.authtoken_to_userid(auth_token))

        return f(*args)
    return wrapper


def encode_exceptions_in_response(f):
    def wrapper(*args):
        self = args[0]

        try:
            return f(*args)
        except Exception as e:

            try:
                ret = e.failure_info_dict()
            except AttributeError:
                ret = dict(
                    error=dict(
                        name=e.__class__.__name__,
                        args=e.args,
                        traceback=traceback.format_tb(e.__traceback__),
                    )
                )

            ret["success"] = False
            self.write(ret)

    return wrapper


class FrameworkWebServiceHandler(tornado.web.RequestHandler):
    keysep = "/"
    def request_body_as_dict(self):
        data = {}
        for k, v in parse_qs(self.request.body).items():
            data[str(k, 'utf-8')] = str(v[0], 'utf-8')
        return data


class CreateHandler(FrameworkWebServiceHandler):

    @encode_exceptions_in_response
    @authed_handler_method
    def put(self, key):
        from kck.lib.kck_data import KCKData
        self.write(KCKData.get_instance().update(key, self.request_body_as_dict()))


class UpdateHandler(FrameworkWebServiceHandler):

    @encode_exceptions_in_response
    @authed_handler_method
    def post(self, key):
        from kck.lib.kck_data import KCKData
        req_body = self.request_body_as_dict()
        result = KCKData.get_instance().update(key, req_body)
        self.write(result)


class FetchHandler(FrameworkWebServiceHandler):
    @encode_exceptions_in_response
    @authed_handler_method
    def get(self, key):

        from kck.lib.kck_cache import KCKCache
        r = KCKCache.get_instance().get(key)
        r["modified"] = str(r["modified"]) if type(r["modified"]) is datetime.datetime else r["modified"]
        self.write(r)


if use_auth:
    class LoginHandler(FrameworkWebServiceHandler):

        @encode_exceptions_in_response
        def post(self, username):
            global use_auth
            global auth_obj
            http_server_instance = KCKHTTPServer.get_instance()
            # auth_obj = _http_server_instance.auth_obj
            passwd = self.request.arguments["password"][0]
            passwd_hash = auth_obj.gen_password_hash(passwd) if use_auth else None

            authkey_template = kck_config_lookup("auth", "auth_user") if use_auth else None
            authkey_elements = authkey_template.split(self.keysep) if authkey_template is not None else None
            key_list = []
            for key_element in authkey_elements:
                if key_element[0] == ":":
                    local_vars = locals()
                    key_list.append(local_vars[key_element[1:]])
                else:
                    key_list.append(key_element)

            # req = self.request
            key = self.keysep.join(key_list)

            from kck.lib.kck_cache import KCKCache
            try:
                r = KCKCache.get_instance().get(key)
            except KCKKeyNotSetException:
                raise AuthLoginUnknownUser(username=username)

            self.write(json.dumps({"success": True, "auth_token": auth_obj.userid_to_authtoken(r["value"]).decode()}))


    class LogoutHandler(FrameworkWebServiceHandler):

        @encode_exceptions_in_response
        @authed_handler_method
        def get(self, username):
            r = self.request


_http_server_instance = None

class KCKHTTPServer(tornado.web.Application):

    # auth_obj = None

    @classmethod
    def get_instance(cls):
        global _http_server_instance

        if _http_server_instance is None:
            _http_server_instance = cls()

        return _http_server_instance

    def __init__(self):
        global use_auth
        routes = [
            (r"/create/(.*)", CreateHandler),
            (r"/update/(.*)", UpdateHandler),
            (r"/fetch/(.*)", FetchHandler),
        ]

        # try:
        #     self.use_auth = kck_config_lookup("kck", "enable_auth")
        # except KeyError:
        #     pass

        if use_auth:
            # self.auth_obj = KCKAuth(auth_dict=kck_config_lookup("auth"))
            # print("adding auth routes")
            routes.append((r"/login/(.*)", LoginHandler),)
            routes.append((r"/logout/(.*)", LogoutHandler), )
        # else:
        #     self.auth_obj = KCKAuth(auth_dict=None)


        super(KCKHTTPServer, self).__init__(
            routes,
            debug=options.debug,
        )




import hashlib

import datetime

import jwt

from kck.lib.config import kck_config_lookup
from kck.lib.exceptions import AuthTokenInvalid, AuthTokenExpired

authed_userid = None
try:
    use_auth = kck_config_lookup("kck", "enable_auth")
except KeyError:
    use_auth = False


class KCKAuth(object):

    auth_secret = kck_config_lookup("auth", "secret") if use_auth else None
    password_hash_salt = kck_config_lookup("auth", "password_hash_salt") if use_auth else None

    def __init__(self, auth_dict):
        self.auth_dict = auth_dict

    def gen_password_hash(self, passwd):
        salt = self.password_hash_salt.encode("utf-8")
        salted_passwd = salt+passwd.encode("utf-8") if type(passwd) is str else salt+passwd
        return hashlib.md5(salted_passwd).hexdigest()

    def type(self):
        return self.auth_dict["type"]

    def userid_to_authtoken(self, user_id):
        currtime = datetime.datetime.utcnow()
        return jwt.encode(
            {
                'exp': currtime + datetime.timedelta(days=0, seconds=5),
                'iat': currtime,
                'sub': user_id
            },
            self.auth_secret,
            algorithm='HS256'
        )

    def authtoken_to_userid(self, token):
        try:
            return jwt.decode(
                token,
                self.auth_secret
            )['sub']
        except jwt.ExpiredSignatureError:
            raise AuthTokenExpired(token)
        except jwt.InvalidTokenError:
            raise AuthTokenInvalid(token)

    def authed_userid(self, userid=None):
        global authed_userid
        if userid is not None:
            authed_userid = userid
        return authed_userid


import os

from .config import kck_config_lookup
from .kck_updater import KCKUpdater
from .yaml_reader import YAMLReader

class KCKYamlUpdater(KCKUpdater, YAMLReader):

    def __init__(self, yaml_filepath):

        self.filepath = yaml_filepath

        self._proc_updater_yaml()

    @classmethod
    def register_updaters(cls, data_obj):
        """register all yaml-based updaters in the updaters dir"""
        # from kck.lib.kck_data import KCKData
        # data_obj = KCKData.get_instance()
        def fnfilter(filepath):
            # print("fp: {}, fpl: {}".format(filepath, filepath[-3:]))
            if filepath[-4:] == ".yml":
                return True
            return False

        def proc_primer_yaml(filepath):
            data_obj.register_updater(
                os.path.basename(filepath)[:-4],
                KCKYamlUpdater(filepath)
            )

        cls.proc_files_in_dirpath(
            cls.make_relpath_absolute(kck_config_lookup("kck", "updaters_dirpath")),
            fnfilter,
            proc_primer_yaml
        )


    def _proc_updater_yaml(self):
        # print("yfp: {}".format(self.filepath))
        u = self.read_yaml(self.filepath)

        # print("u: {}".format(u))

        # the key is the same as the yaml file (minus the .yml)
        self.name = [os.path.basename(self.filepath)[:-4]]

        self.database = u["database"]

        self.parameters = u["parameters"]
        # print("parameters: {}".format(self.parameters))

        # self.expire = datetime.timedelta(seconds=u["expire_seconds"])

        self.query_template = u["query"]["template"]



import os
import yaml

from .err import handle_error

SCRIPTDIRPATH = os.path.dirname(os.path.realpath(__file__))

_conf_obj = None

def get_kck_environment():
    return "dev" \
        if "KCK_ENVIRONMENT" not in os.environ \
            or os.environ["KCK_ENVIRONMENT"] not in ["dev", "prod", "test", "test-jwt", "test-refresh"] \
        else os.environ["KCK_ENVIRONMENT"]

def kck_config_lookup(*args, **kwargs):
    on_error = kwargs["on_error"] if "on_error" in kwargs else None
    conf_ptr = _get_config_object(on_error=on_error)

    if args is not None:
        for arg in args:
            conf_ptr = conf_ptr[arg]
    return conf_ptr

def _kck_config_filepath():
    if "KCK_CONFIG" in os.environ:
        return os.environ["KCK_CONFIG"]
    return os.path.join(SCRIPTDIRPATH, "..", "..", "config", "{}.yml".format(get_kck_environment()))

def _get_config_object(on_error=None):
    global _conf_obj
    if _conf_obj is None:
        with open(_kck_config_filepath(), 'r') as stream:
            try:
                _conf_obj = yaml.load(stream)
            except yaml.YAMLError as exc:
                handle_error(exc, "unable to init config object", on_error)
    return _conf_obj




from kck.lib.kck_cache import KCKCache
from kck.lib.kck_primer import KCKPrimer
import types

def NoDataAvailable(Exception):
    key = None

    def __init__(self, key):
        self.key = key

def get_decorator_primer(keys, compute_func):

    def compute_with_self_and_args(self, keystr):
        args = keystr.split(KCKCache.keysep)[1:]
        return compute_func(*args)

    p = KCKPrimer()
    p.keys = keys
    p.compute = types.MethodType(compute_with_self_and_args, p)
    return p

class kckprime:

    dargs = None
    dkwargs = None

    def __init__(self, *args, **kwargs):
        self.dargs = args
        self.dkwargs = kwargs

    def __call__(self, f, *args, **kwargs):

        def cache_lookup(*args, **kwargs):
            cache_obj = KCKCache.get_instance()
            key = "{}{}{}".format(self.dkwargs["name"],cache_obj.keysep,cache_obj.keysep.join(args))
            response = cache_obj.get(key)
            if bool(response["success"]):
                return response["value"]
            raise NoDataAvailable(key)

        def df(*args, **kwargs):
            # set up primer
            keys = self.dkwargs["keys"]
            name = self.dkwargs["name"]

            cache_obj = KCKCache.get_instance()
            if not cache_obj.primer_registered(name):
                cache_obj.register_primer(name, get_decorator_primer(keys, f))

            return cache_lookup(*args, **kwargs)

        return df

class kckrefresh:
    dargs = None
    dkwargs = None

    def __init__(self, *args, **kwargs):
        self.dargs = args
        self.dkwargs = kwargs

    def __call__(self, f, *args, **kwargs):
        def df(*args, **kwargs):
            ret =  f(*args, **kwargs)

            param_dict = {}
            for compute_param_name, param_name in self.dkwargs["param_keys"].items():
                param_dict[compute_param_name] = kwargs[param_name]

            key_param_list = []
            for param_name in self.dkwargs["keys"]:
                key_param_list.append(param_dict[param_name])

            cache_obj = KCKCache.get_instance()
            name = self.dkwargs["name"]
            keys = self.dkwargs["keys"]
            if not cache_obj.primer_registered(name):
                cache_obj.register_primer(name, get_decorator_primer(keys, self.dkwargs["compute"]))

            refresh_key = "{}{}{}".format(self.dkwargs["name"], cache_obj.keysep, cache_obj.keysep.join(key_param_list))
            cache_obj.refresh(refresh_key)

            return ret

        return df



from .log import log
from .cassandra_actor import CassandraActor

process_data_singleton = None

class KCKData(CassandraActor):

    _pcol = None
    parameters = None
    cache_obj = None

    @classmethod
    def get_instance(cls, **kwargs):
        global process_data_singleton
        if process_data_singleton is None:
            process_data_singleton = cls(**kwargs)
        return process_data_singleton

    def __init__(self, inhibit_framework_registration=False, cache_obj=None):
        self.primitive_init()
        if cache_obj is None:
            from kck.lib.kck_cache import KCKCache
            self.init(inhibit_framework_registration=inhibit_framework_registration, cache_obj=KCKCache.get_instance(data_obj=self), data_obj=self)
        else:
            self.init(inhibit_framework_registration=inhibit_framework_registration, cache_obj=cache_obj, data_obj=self)

    def register_updater(self, name, updater_obj):
        updater_obj.set_data(self)
        updater_obj.set_cache(self.cache_obj)
        updater_obj.set_name(name)
        self.updaters[name] = updater_obj

    def _updater_obj(self, key):
        ret = self.updater(key.split(self.keysep)[0])
        return ret

    def updater(self, name):
        return self.updaters[name]

    def update(self, keystr, data):
        u = self._updater_obj(keystr)
        u.do_hooks("pre_update", key=keystr, hints=data)
        key_data = u.key_to_param_dict(key_str=keystr)
        key_data.update(data)
        r = u.update(key_data)
        u.do_hooks("post_update", key=keystr, hints=data)
        return dict(success=True)
